from sqlalchemy import event, text
from sqlalchemy.engine import Engine
from sqlalchemy.sql.selectable import Select, Alias


class SuaveDeleteMixin:
    __use_soft_deletes__ = True


def is_soft_delete_supported(instance):
    return getattr(instance, "__use_soft_deletes__", False) and hasattr(instance, "deleted_at")


@event.listens_for(Engine, "before_execute", retval=True)
def before_execute_listener(conn, clause, multi_params, params):
    if (not isinstance(clause, Select) or
            not hasattr(clause, "froms") or
            not isinstance(clause.froms, list)):
        return clause, multi_params, params

    for from_clause in clause.froms:

        if isinstance(from_clause, Alias):
            continue

        table_name = from_clause.name
        filter_condition = text(f"{table_name}.deleted_at IS NULL")
        clause = clause.where(filter_condition)

    return clause, multi_params, params





from datetime import datetime, UTC

from sqlalchemy import and_
from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session, class_mapper, sessionmaker

from suave_deletes.mixins import is_soft_delete_supported
from suave_deletes.queries import SuaveDeleteQuery


class SuaveDeleteSession(Session):
    def query(self, *entities, **kwargs):
        query = super().query(*entities, **kwargs)

        for instance in entities:
            if is_soft_delete_supported(instance):
                mapper = class_mapper(instance)
                query = query.filter(and_(mapper.c.deleted_at == None))  # noqa: E711

        return query

    def delete(self, instance):
        if is_soft_delete_supported(instance):
            instance.deleted_at = datetime.now(UTC)
        else:
            super().delete(instance)

        self._cascade_delete(instance)

    def _cascade_delete(self, instance):
        mapper = class_mapper(instance.__class__)
        relationships = mapper.relationships

        for relationship in relationships:
            cascade = relationship.cascade

            if "delete" not in cascade:
                continue

            related_instances = getattr(instance, relationship.key)

            if related_instances is None:
                continue

            if isinstance(related_instances, list):
                for related_instance in related_instances:
                    self.delete(related_instance)
            else:
                self.delete(related_instances)


def create_suave_delete_session(engine: Engine) -> sessionmaker[SuaveDeleteSession]:
    return sessionmaker(bind=engine, class_=SuaveDeleteSession, query_cls=SuaveDeleteQuery)



from sqlalchemy import and_
from sqlalchemy.orm import Query, class_mapper
from sqlalchemy.sql.elements import Null

from suave_deletes.mixins import is_soft_delete_supported


class SuaveDeleteQuery(Query):
    def join(self, *entities, **kwargs) -> "SuaveDeleteQuery":
        query = super().join(*entities, **kwargs)

        for instance in entities:
            if is_soft_delete_supported(instance):
                mapper = class_mapper(instance)
                query = query.filter(and_(mapper.c.deleted_at == None))  # noqa: E711

        return query

    def with_deleted_at(self) -> "SuaveDeleteQuery":
        if not self._where_criteria:
            return self

        self._where_criteria = tuple(
            c for c in self._where_criteria if not self._has_deleted_at_filter(c)
        )

        return self

    def _has_deleted_at_filter(self, clause) -> bool:
        return (
            clause.left.name == "deleted_at"
            and clause.operator.__name__ == "is_"
            and isinstance(clause.right, Null)
        )



import pytest

from sqlalchemy import Column, Integer, String, DateTime, create_engine
from sqlalchemy.orm import relationship, declarative_base
from sqlalchemy import ForeignKey

from suave_deletes.mixins import SuaveDeleteMixin
from suave_deletes.sessions import create_suave_delete_session

Base = declarative_base()


class User(SuaveDeleteMixin, Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    deleted_at = Column(DateTime)


class Workspace(SuaveDeleteMixin, Base):
    __tablename__ = 'workspaces'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    participants = relationship(
        "Participant",
        back_populates="workspace",
        cascade="all, delete, delete-orphan"
    )
    deleted_at = Column(DateTime)


class Participant(SuaveDeleteMixin, Base):
    __tablename__ = 'participants'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    workspace_id = Column(Integer, ForeignKey('workspaces.id'))
    workspace = relationship("Workspace", back_populates="participants")
    deleted_at = Column(DateTime, nullable=True)


@pytest.fixture(scope='module')
def engine():
    return create_engine('sqlite:///:memory:')


@pytest.fixture(scope='module')
def tables(engine):
    Base.metadata.create_all(engine)
    yield
    Base.metadata.drop_all(engine)

@pytest.fixture
def session(engine, tables):
    Session = create_suave_delete_session(engine)
    session = Session()
    yield session
    session.close()

def test_soft_delete(session):
    user = User(name="Test User")
    session.add(user)
    session.commit()

    assert session.query(User).count() == 1

    session.delete(user)
    session.commit()

    assert session.query(User).count() == 0

    assert session.query(User).with_deleted_at().count() == 1


def test_query_with_soft_delete(session):
    user_john = User(name="John")
    user_karen = User(name="Karen")

    session.add(user_john)
    session.add(user_karen)
    session.commit()

    assert session.query(User).count() == 2

    session.delete(user_karen)
    session.commit()

    users = session.query(User).all()

    assert len(users) == 1
    assert users[0].name == "John"


def test_soft_delete_on_participants(session):
    workspace = Workspace(name="Test Workspace")
    participant1 = Participant(name="Participant 1", workspace=workspace)
    participant2 = Participant(name="Participant 2", workspace=workspace)

    workspace.participants.append(participant1)
    workspace.participants.append(participant2)

    session.add(workspace)
    session.commit()

    assert len(workspace.participants) == 2

    session.delete(participant1)
    session.commit()

    assert len(workspace.participants) == 1
    assert workspace.participants[0].name == "Participant 2"



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="suave_deletes",
    version="0.1.1",
    packages=find_packages(),
    author="Stefan Framework",
    description="A simple soft delete implementation for sqlalchemy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/stefanFramework/suave_deletes",
    install_requires=[
        'SQLAlchemy>=2.0.0',
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)